//    XXML/Xerces Parser
//    Copyright (C) 2005 Robert S. Thau
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License verstion 2.1.
//
//    This library is distributed as a proof-of-concept only,
//    and WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package org.rsthau.xxml;

import org.xml.sax.InputSource;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.sax.SAXSource;  
import javax.xml.transform.stream.StreamResult; 

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;

import org.xml.sax.XMLReader;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.ErrorHandler;

import java.io.*;

public class TestDriver {

    static boolean filesSame (File refFile, File outFile)
	throws IOException
    {
	if (!refFile.exists()) {
	    System.err.println ("ERROR: Not found: " + refFile);
	    return false;
	}

	BufferedReader refReader= new BufferedReader(new FileReader (refFile));
	BufferedReader outReader= new BufferedReader(new FileReader (outFile));
	int line = 0;
	
	while (true) {
	    String refLine = refReader.readLine();
	    String outLine = outReader.readLine();
	    ++line;

	    if (refLine == null && outLine == null) return true;

	    if (refLine == null) {
		System.out.println (outFile.toString() + 
				    " has extra lines starting at " + line);
		return false;
	    }

	    if (outLine == null) {
		System.out.println (outFile.toString() + 
				    " truncated at line " + (line - 1));
		return false;
	    }

	    if (!refLine.equals(outLine)) {
		System.out.println (outFile.toString() + " differs from " +
				    refFile.toString() + " at line " + line);
		System.out.println (outFile.toString() + ": " + outLine);
		System.out.println (refFile.toString() + ": " + refLine);
		return false;
	    }
	}
    }

    static boolean smokeTestDom (String dirName, String testName)
	throws IOException
    {
	try {
	    File inFile  = new File (dirName, testName + ".in");
	    File errs    = new File (dirName, testName + ".verr");
	    File errsChk = new File (dirName, testName + ".vchk");

	    DocumentBuilderFactory factory = 
		DocumentBuilderFactory.newInstance();

	    // For DOM, test validation as well...

	    factory.setValidating(true);

	    DocumentBuilder builder = factory.newDocumentBuilder();
	    TestErrorHandler errh = new TestErrorHandler (errs, errsChk);
	    builder.setErrorHandler (errh);

	    try {
		Document doc = builder.parse(inFile);
	    }
	    catch (SAXException e) {
		// We trust the error handler will have been called...
	    }

	    errh.wrapUp();
	    return errh.isOk();
	} 
	catch (ParserConfigurationException e) {
	    return false;
	} 
    }

    static boolean processFile (String dirName, String testName)
	throws IOException
    {
	try {
	    File inFile  = new File (dirName, testName + ".in");
	    File refFile = new File (dirName, testName + ".chk");
	    File outFile = new File (dirName, testName + ".out");
	    File errs    = new File (dirName, testName + ".errs");
	    File errsChk = new File (dirName, testName + ".echk");

	    TransformerFactory tFactory = TransformerFactory.newInstance();
	    Transformer transformer = tFactory.newTransformer();

	    Writer writer = new BufferedWriter (new FileWriter (outFile));
	    InputSource isource = new InputSource (inFile.toString());

	    XMLReader parser = new XxmlSaxParser();
	    TestErrorHandler errh = new TestErrorHandler (errs, errsChk);
	    parser.setErrorHandler (errh);

	    SAXSource source = new SAXSource (parser, isource);
	    TransformerException abend = null;

	    try {
		transformer.transform (source, new StreamResult (writer));
	    }
	    catch (TransformerException e) {
		abend = e;
	    }

	    writer.close();
	    errh.wrapUp();

	    boolean success = true;
	    success = errh.isOk()                      && success;
	    success = filesSame (refFile, outFile)     && success;
	    success = smokeTestDom (dirName, testName) && success;

	    if (success) {
		System.out.println ("Test " + inFile + " succeeded");
		outFile.delete();
		return true;
	    }
	    else {
		System.out.println ("FAILURE: Test " + inFile);
		if (abend != null)
		    System.out.println ("Parse ends on: "+abend.getMessage());
		return false;
	    }
	}
	catch (IOException e) {
	    throw e;
	}
	catch (Exception e) {
	    e.printStackTrace (System.out);
	    return false;
	}
    }

    static boolean processDir (String dirName) 
	throws IOException
    {
	int tests_run = 0;
	int tests_failed = 0;

	for (int i = 1; ; ++i) {
	    String testName = "test" + i;
	    File inFile  = new File (dirName, testName + ".in");

	    if (!inFile.exists()) {
		if (i == 1) 
		    throw new RuntimeException
			("Could not find test cases in " + dirName);
		break;
	    }

	    System.out.println ("Trying " + inFile);
	    ++tests_run;
	    if (!processFile (dirName, testName)) {
		++tests_failed;
		System.out.println ("Test " + inFile + " failed");
		// break;
	    }
	}

	System.err.println ("Ran " + tests_run + " tests; " 
			    + tests_failed + " failures.");

	return tests_failed == 0;
    }

    public static void main (String args[])
	throws Exception
    {
	if (args.length == 1)
	    System.exit (processDir (args[0]) ? 0 : 1);
	else {
	    String dirName = args[0];

	    for (int i = 1; i < args.length; ++i) {
		System.out.println ("Trying " + args[i]);
		processFile (dirName, args[i]);
	    }
	}
    }
}
