//    XXML/Xerces Parser
//    Copyright (C) 2005 Robert S. Thau
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License verstion 2.1.
//
//    This library is distributed as a proof-of-concept only,
//    and WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package org.rsthau.xxml;

class Token {

    // Token types:

    public final static int EOF        = -1;
    public final static int NAME       = 0;
    public final static int TEXT       = 1;
    public final static int EQUALS     = 2;
    public final static int KEYWORD    = 3;
    public final static int COMMENT    = 4;
    public final static int COLON      = 5;
    public final static int LBRACE     = 6;
    public final static int RBRACE     = 7;
    public final static int SEMICOLON  = 8;

    // Notional tokens -- BEGIN_FILE is logical whitespace that
    // gets pushed on the indentation stack so the corresponding
    // EOF has something to match; BODY_END is a semicolon or
    // terminating '}'.

    public final static int BEGIN_FILE = 9;
    public final static int BODY_END   = 10;

    // Fields of an individual token.
    // NB these are immutable objects...

    public final int type;
    public final int indentLevel;
    public final Token match;
    public final String str;
    public final boolean macroExpanded;

    public Token (int type, int indentLevel) {
	this.type = type;
	this.indentLevel = indentLevel;
	this.match = null;
	this.macroExpanded = false;
	switch (type) {
	case COLON:     str = ":"; break;
	case EQUALS:    str = "="; break;
	case LBRACE:    str = "{"; break;
	case RBRACE:    str = "}"; break;
	case SEMICOLON: str = ";"; break;
	case EOF:       str = "EOF"; break;
	default: throw new RuntimeException 
		("Can't create token type " + type
		 + " without source text");
	}
    }

    public Token (int type, Token match) {
	// This should be an assert, but JDK 1.4 asserts are useless.
	if (type != BODY_END)
	    throw new RuntimeException
		("Internal screw-up:  someone didn't understand the API");

	this.type = type;
	this.indentLevel = match.indentLevel;
	this.match = match;
	this.str = "[end " + match.str + "]";
	this.macroExpanded = false;
    }

    public Token (int type, String str, int indentLevel) {
	this.type = type;
	this.str = str;
	this.match = null;
	this.indentLevel = indentLevel;
	this.macroExpanded = false;
    }

    // This constructor used in macroexpansion...

    public Token (Token other, int indentLevel) {
	this.type  = other.type;
	this.str   = other.str;
	this.match = null;
	this.indentLevel = indentLevel;
	this.macroExpanded = true;
    }

    public String toString () {
	return "[" + indentLevel + "," + type + 
	    (macroExpanded ? "m" : "") +
	    "] '" + str + "'";
    }

    // Indentation syntax handling ... I'm trying to be a bit declarative
    // about what ends a group started by which other tokens:

    public boolean delimitsIndentRegion () {
	return this.type == LBRACE || this.type == BEGIN_FILE; 
    }
    
    // This method is called by the tokenizer/group manager when trying to
    // manage the grouping stack...

    public boolean endedBy (Token other) 
	throws SyntaxError
    {
	if (this.type == LBRACE) {
	    if (other.type == RBRACE) return true;
	    if (other.type == EOF || other.type == SEMICOLON)
		throw new SyntaxError ("Unterminated '{'");
	    return false;
	}

	if (this.type == BEGIN_FILE)
	    return (other.type == EOF);

	if (other.type == RBRACE 
	    || other.type == SEMICOLON 
	    || other.type == EOF)
	    return true;

	// Monkey business here:  all tokens of a macroexpansion
	// are put at the same logical position as the token that
	// triggered the macroexpansion.  So, they'd all terminate
	// each other if we did indentation normally.  Instead,
	// the macroexpander puts explicit terminators in, and
	// we suppress the indentation machinery like so:

	if (this.macroExpanded && other.macroExpanded)
	    return false;

	if (this.indentLevel >= other.indentLevel) {

	    // MONDO kludge-o:  require explicit terminator on
	    // \doctype and \include, to avoid weirdness when the
	    // XMLEntityScanner swallows whitespace after the
	    // end of an external entity (bad behavior that I
	    // haven't fully traced)...

	    if (this.type == KEYWORD &&
		(this.str.equals ("include") || this.str.equals ("doctype")))
		throw new SyntaxError ("\\" + this.str + " must end with ';'");

	    return true;
	}

	return false;
    }

    public boolean explicitEnd () {
	return type == RBRACE || type == EOF || type == SEMICOLON;
    }

    public boolean explicitEnd (Token other)
    {
	if (this.type == RBRACE)    return (other.type == LBRACE);
	if (this.type == EOF)       return (other.type == BEGIN_FILE);
	if (this.type == SEMICOLON) return true;
	return false;
    }
}
