//    XXML/Xerces Parser
//    Copyright (C) 2005 Robert S. Thau
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License verstion 2.1.
//
//    This library is distributed as a proof-of-concept only,
//    and WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package org.rsthau.xxml;

import java.io.IOException;

import java.util.Locale;
import java.util.Stack;
import java.util.Vector;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Enumeration;

// These are supposed to be documented:

import org.apache.xerces.xni.XMLDocumentHandler;
import org.apache.xerces.xni.XMLString;
import org.apache.xerces.xni.XNIException;
import org.apache.xerces.xni.XMLAttributes;
import org.apache.xerces.xni.XMLLocator;
import org.apache.xerces.xni.Augmentations;
import org.apache.xerces.xni.QName;

import org.apache.xerces.xni.parser.XMLDocumentSource;
import org.apache.xerces.xni.parser.XMLDTDScanner;

// And these probably as well:

import org.apache.xerces.util.XMLAttributesImpl;
import org.apache.xerces.util.XMLSymbols;
import org.apache.xerces.util.AugmentationsImpl;
import org.apache.xerces.util.NamespaceSupport;
import org.apache.xerces.util.MessageFormatter;

// But at this point, I'm probably wandering off into high weeds:

import org.apache.xerces.impl.XMLEntityManager;
import org.apache.xerces.impl.XMLErrorReporter;

class XxmlScanner 
    implements XMLDocumentSource
{
    private XMLDocumentHandler contentHandler_;
    private XMLLocator locator_;
    private XMLEntityManager entityManager_;
    private XMLErrorReporter errorReporter_;
    private Augmentations augs = new AugmentationsImpl();

    private XxmlTokenizer tokenizer_;

    private boolean seenRoot_ = false;
    private boolean seenDoctype_ = false;

    private Hashtable posAttrs_    = new Hashtable();
    private Hashtable stringExpns_ = new Hashtable();

    // Setup

    XxmlScanner (XMLEntityManager em, 
		 XMLDocumentHandler ds, 
		 XMLDTDScanner dts,
		 XMLErrorReporter er) 
	throws XNIException, IOException, SyntaxError
    {
	contentHandler_ = ds;
	entityManager_  = em;
	ds.setDocumentSource (this);
	locator_ = em.getEntityScanner();
	tokenizer_ = new XxmlTokenizer (em.getEntityScanner(), em, dts);
	em.setEntityHandler (tokenizer_);
	indentStack_.push (new Token (Token.BEGIN_FILE, "[xxml]", 
				      XxmlTokenizer.EOF_INDENT));

	if (contentHandler_ == null)
	    throw new XNIException ("Content handler unset");

	contentHandler_.startDocument(locator_, "UTF-8", 
				      new NamespaceSupport(), augs);

	errorReporter_ = er;

	er.putMessageFormatter 
	    ("org.rsthau.xxml",
	     new MessageFormatter() {

		 // They've actually got the full I18N machinery in
		 // org.apache.xerces.impl.msg.XMLMessageFormatter -- 
		 // and messages in only one language, English.  Sigh...

		 public String formatMessage (Locale locale,
					      String key,
					      Object[] args)
		 {
		     try {
			 return java.text.MessageFormat.format (key, args);
		     }
		     catch (Exception ex) {
			 return key;
		     }
		 }
	     });
    }

    // Data structures for parsing.

    private Token prev_tok_;
    private Token cur_tok_;

    private Stack indentStack_ = new Stack();
    private LinkedList deferredToks_ = new LinkedList();

    // Read the next "parse-relevant" token, taking note of grouping
    // (generating BODY_END tokens where needed), and skipping over
    // logical whitespace (comments, curly-braces, etc.).

    public Token nextToken()
	throws IOException, SyntaxError
    {
	prev_tok_ = cur_tok_;

	do {
	    if (cur_tok_ != null && cur_tok_.type == Token.EOF)
		throw new RuntimeException ("Internal error:  read past EOF");

	    // Read next token (or reread it, if we deferred one
	    // for an indentation-generated BODY_END).

	    if (deferredToks_.isEmpty()) 
		cur_tok_ = tokenizer_.next();
	    else
		cur_tok_ = (Token)deferredToks_.removeFirst();

	    // Also peek at the top of the indent stack...

	    Token innermost_open = (Token)indentStack_.peek(); 

	    // Special treatment here for some token types.
	    // Turn \enddocument into logical EOF

	    if (cur_tok_.type == Token.KEYWORD
		&& cur_tok_.str.equals ("enddocument"))
		cur_tok_ = new Token (Token.EOF, XxmlTokenizer.EOF_INDENT);

	    // LBRACE gets pushed onto the indent stack...
	    // on *top* of the element that will actually be
	    // terminated by the matching '}'.

	    if (cur_tok_.type == Token.LBRACE) {

		if (prev_tok_.type == Token.KEYWORD) {
		    if (!prev_tok_.str.equals ("each_string"))
			throw new SyntaxError ("Misplaced '{'");
		}
		else if (prev_tok_.type != Token.COLON)
		    throw new SyntaxError ("Misplaced '{'");

		prev_tok_ = cur_tok_; // Catch '{{'
		indentStack_.push (cur_tok_);
	    }

	    else if (cur_tok_.type == Token.BEGIN_FILE) {
		indentStack_.push (cur_tok_);
	    }

	    else if (innermost_open.endedBy (cur_tok_)) {

		// We may want to rescan this token, e.g.,
		// if it terminates a current indentation group and
		// begins a new one in its own right.

		if (!cur_tok_.explicitEnd (innermost_open))
		    deferredToks_.addFirst (cur_tok_);

		// Terminating what's on the top of the stack.  Pop it off... 

		indentStack_.pop();

		// Ugly hack... if we are terminating a '{', we actually
		// mean to terminate the element *underneath*...

		if (innermost_open.type == Token.LBRACE)
		    innermost_open = (Token)indentStack_.pop();

		cur_tok_ = new Token (Token.BODY_END, innermost_open);
	    }
	    else if (cur_tok_.explicitEnd()) {
		throw new SyntaxError(cur_tok_.str + 
				      " with nothing to terminate");
	    }
	} while (!indentStack_.empty() && logicalWhitespace (cur_tok_));
	
	if (indentStack_.empty()) 
	    cur_tok_ = new Token (Token.EOF, XxmlTokenizer.EOF_INDENT);

	return cur_tok_;
    }
    
    private boolean logicalWhitespace (Token tok) {

	// Curly braces and delimitation of include files are
	// invisible to the parser.

	return 
	    tok.type == Token.COMMENT
	    || (tok.type == Token.LBRACE)
	    || (tok.type == Token.BEGIN_FILE)
	    || (tok.type == Token.BODY_END
		&& (tok.match.type == Token.LBRACE
		    || tok.match.type == Token.BEGIN_FILE));
    }

    // Parsing.

    public boolean parse (boolean scanAll)
	throws IOException, XNIException, SyntaxError
    {
	if (cur_tok_ != null && cur_tok_.type == Token.EOF) return false;

	do {
	    try {
		switch (nextToken().type) {

		case Token.EOF:      break;
		case Token.NAME:     parseElement (cur_tok_); break;
		case Token.KEYWORD:  
		    if (cur_tok_.str.equals ("bulk_text"))
			doBulkText();
		    else
			parseDirective (cur_tok_); 
		    break;

		case Token.BODY_END: 
		    if (cur_tok_.match.type != Token.NAME)
			throw new XNIException
			    ("Desynchronized indent stack:  terminating "
			     + cur_tok_.match.str);
		    endElement (cur_tok_.match.str); 
		    break;

		case Token.TEXT:
		    if (!specialTextSyntax ()) {

			// Default case -- assume text or mixed content model.

			String s = cur_tok_.str;

			if (prev_tok_ != null && prev_tok_.type == Token.TEXT)
			    s = " " + s;

			XMLString xmls = new XMLString (s.toCharArray(), 0, 
							s.length());
			contentHandler_.characters(xmls, augs);
		    }
		    break;

		default: 
		    throw new SyntaxError ("Unexpected " + cur_tok_.str);
		}
	    }
	    catch (SyntaxError syn) {
		errorReporter_.reportError 
		    ("org.rsthau.xxml", syn.getMessage(), new Object [0],
		     XMLErrorReporter.SEVERITY_FATAL_ERROR);
	    }
	} while (scanAll && cur_tok_.type != Token.EOF);

	if (cur_tok_.type != Token.EOF) { 
	    return true; 
	}
	else {
	    XMLString newline = new XMLString ("\n".toCharArray(), 
					       0, 1);
	    contentHandler_.ignorableWhitespace (newline, augs);
	    contentHandler_.endDocument(augs);
	    return false;
	}
    }

    private void parseElement (Token nameTok)
	throws IOException, SyntaxError, XNIException
    {
	seenRoot_ = true;
	indentStack_.push (nameTok);

	// Scan for attributes...

	XMLAttributesImpl attrs = new XMLAttributesImpl();
	Vector posAttrs = (Vector) posAttrs_.get (nameTok.str);
	int posAttrCount = 0;
	QName attrName;

	while (true) {
	    Token tok = nextToken();
	    switch (tok.type) {
	    case Token.NAME:
		posAttrCount = -1; // No positional attrs after a named one
		attrName = qnameize (tok.str);
		consume (Token.EQUALS, "'='");
		attrs.addAttribute (attrName, XMLSymbols.fCDATASymbol, 
				    consume(Token.TEXT,"Attribute value").str);
		break;

	    case Token.TEXT:
		if (posAttrCount < 0) {
		    throw new SyntaxError 
			("Unexpected \"" + tok.str + "\";  " + 
			 "No positional attributes after a named one");
		}
		if (posAttrs == null)
		    throw new SyntaxError
			("No positional attributes declared for "+nameTok.str);
		if (posAttrCount >= posAttrs.size())
		    throw new SyntaxError
			("Too may positional attributes for " + nameTok.str);
		attrName = (QName)posAttrs.elementAt (posAttrCount++);
		attrs.addAttribute (attrName, "CDATA", cur_tok_.str);
		break;

	    case Token.COLON: 
		startElement (nameTok.str, attrs);
		return;

	    case Token.BODY_END:
		if (tok.match != nameTok)
		    throw new XNIException
			("Desynchronized element stack in empty element");

		startElement (nameTok.str, attrs);
		endElement (nameTok.str);
		return;

	    default:
		throw new SyntaxError ("Unexpected " + tok.str);
	    }
	}
    }

    private void parseDirective (Token current)
	throws IOException, XNIException, SyntaxError
    {
	indentStack_.push (current);
	if (current.str.equals ("doctype"))
	    scanDocType();
	else if (current.str.equals ("include"))
	    doInclude();
	else if (current.str.equals ("syntax"))
	    doSyntax();
	else
	    throw new SyntaxError ("Unrecognized keyword \\" + current.str);
    }

    private void scanDocType()
	throws IOException, XNIException, SyntaxError
    {
	if (seenRoot_) throw new XNIException ("\\doctype after content");

	String rootElt = consume (Token.NAME, "root document element").str;
	String pubsys  = consume (Token.NAME, "PUBLIC or SYSTEM").str;

	String sysId = null, pubId = null;

	if (pubsys.equals ("SYSTEM"))
	    sysId = consume (Token.TEXT, "System ID").str;
	else if (pubsys.equals ("PUBLIC")) {
	    pubId = consume (Token.TEXT, "Public ID").str;
	    sysId = consume (Token.TEXT, "System ID").str;
	}
	else
	    throw new SyntaxError ("\\doctype must be followed by " +
				   "PUBLIC or SYSTEM");

	consume (Token.BODY_END, "End of DTD declaration");

	contentHandler_.doctypeDecl (rootElt, pubId, sysId, augs);
	tokenizer_.scanExternalDtd (pubId, sysId);
    }

    private void doBulkText()
	throws SyntaxError, IOException, XNIException
    {
	Token elt = currentEnclosingElt();
	if (elt == null)
	    throw new SyntaxError ("Bulk text outside root element");
	
	String theText = tokenizer_.scanBulkText (elt.indentLevel);
	Token textTok = new Token (Token.TEXT, theText, elt.indentLevel + 1);
	deferredToks_.addFirst (textTok);
    }

    private void doInclude()
	throws IOException, XNIException, SyntaxError
    {
	String include_name;
	boolean isCodeResource = false;

	switch (nextToken().type) {
	case Token.TEXT: include_name = cur_tok_.str; break;

	case Token.NAME: 
	    if (cur_tok_.str.equals ("file")) {
		include_name = consume (Token.TEXT, "File name").str;
		break;
	    }
	    else if (cur_tok_.str.equals ("code_resource")) {
		include_name = consume (Token.TEXT, "Code resource name").str;
		isCodeResource = true;
		break;
	    }
	    else throw new SyntaxError ("\\include type '" + cur_tok_.str +
					"' not supported");

	default:
	    throw new SyntaxError ("Unexpected " + cur_tok_.str + 
				   " after \\include");
	}

	include_name = include_name.intern();
	consume (Token.BODY_END, "End of \\include directive");

	if (isCodeResource) {
	    // \include code_resource.

	    // BUG:  Not clear which classLoader to use, but it won't always
	    // be ours.  We could use SecurityManager methods to walk
	    // the stack and try to get a better idea...

	    ClassLoader cloader = this.getClass().getClassLoader();
	    java.net.URL theResource = cloader.getResource (include_name);
	    if (theResource == null) {
		throw new SyntaxError ("Could not locate code_resource " +
				       include_name);
	    }
	    else {
		include_name = theResource.toString();
	    }
	}

	String entityName = include_name; // to catch recursive \includes
	entityManager_.addExternalEntity (entityName, null, 
					  include_name, null);
	entityManager_.startEntity (entityName, false);
    }

    private Token consume (int tokType, String whatWeExpect) 
	throws IOException, SyntaxError
    {
	Token current = nextToken();

	if (current.type != tokType)
	    throw new SyntaxError ("Expected " + whatWeExpect + 
				   "; got " + current.str);

	return current;
    }

    private Token currentEnclosingElt() {
	Token elt = null;

	for (int i = indentStack_.size() - 1; i >= 0; --i) {
	    Token it = (Token)indentStack_.elementAt (i);
	    if (it.type == Token.NAME) {
		elt = it;
		break;
	    }
	}

	return elt;
    }

    // Magic syntax declarations, and some of the handling...

    private void doSyntax()
	throws IOException, XNIException, SyntaxError
    {
	String targetName = consume (Token.NAME, "Element name").str;
	consume (Token.COLON, ":");

	while (nextToken().type != Token.EOF) {
	    switch (cur_tok_.type) {
	    case Token.BODY_END: return;
	    case Token.KEYWORD: 
		indentStack_.push (cur_tok_);
		scanSyntaxClause (targetName, cur_tok_); 
		break;
	    default:
		throw new SyntaxError
		    ("Can't begin a syntax clause with " + cur_tok_.str);
	    }
	}
    }

    private void scanSyntaxClause (String tagName, Token clauseTok) 
	throws IOException, XNIException, SyntaxError
    {
	if (clauseTok.str.equals("posattrs")) {
	    Vector posattrs = new Vector();
	    posAttrs_.put (tagName, posattrs);
	    while (nextToken().type != Token.EOF) {
		switch (cur_tok_.type) {
		case Token.BODY_END: return;
		case Token.NAME: posattrs.add (qnameize (cur_tok_.str)); break;
		default:
		    throw new SyntaxError
			("'" + cur_tok_.str + "' in posattrs declaration");
		}
	    }

	    throw new SyntaxError ("EOF in \\syntax \\posattrs");
	}
	else if (clauseTok.str.equals("each_string")) {
	    LinkedList each_string = new LinkedList();
	    stringExpns_.put (tagName, each_string);

	    // NB we put tokens in each_string in reverse order, 
	    // because they'll be implicitly reversed again when
	    // specialTextSyntax pushes them onto the front of
	    // deferredToks_.

	    while (nextToken().type != Token.EOF) {
		if (cur_tok_.type == Token.BODY_END) {
		    if (cur_tok_.match != clauseTok) 
			each_string.addFirst (new Token (Token.SEMICOLON, 0));
		    else {
			// Want generated tags terminated by the
			// terminator in the source text, not the
			// one in the macroexpansion...

			if (each_string.size() == 0)
			    throw new SyntaxError ("Empty \\each_string");
			else if (((Token)each_string.getFirst()).type
				 == Token.SEMICOLON)
			    each_string.removeFirst();

			return;
		    }
		}
		else if (cur_tok_.type == Token.NAME) {

		    // We have to mimic operation of the indent stack
		    // while collecting tokens so that BODY_ENDs will
		    // show up in the right places.

		    if (each_string.isEmpty() 
			|| prev_tok_.type == Token.COLON
			|| prev_tok_.type == Token.LBRACE)
		    {
			indentStack_.push (cur_tok_);
		    }

		    each_string.addFirst (cur_tok_);
		}
		else {
		    if (cur_tok_.type == Token.KEYWORD) {
			if (!cur_tok_.str.equals ("it"))
			    throw new SyntaxError
				("Unexpected keyword \\" + cur_tok_.str);
		    }

		    each_string.addFirst (cur_tok_);
		}
	    }

	    throw new SyntaxError ("EOF in \\syntax \\each_string");
	}
	else 
	    throw new SyntaxError ("Unrecognized syntax clause: " + 
				   clauseTok.str);
    }

    private boolean specialTextSyntax ()
	throws SyntaxError, XNIException, IOException
    {
	// Identify enclosing element, if any...

	Token elt = currentEnclosingElt();
	if (elt == null) return false;

	// Does it have an \each_string?

	LinkedList expn = (LinkedList) stringExpns_.get (elt.str);

	if (expn == null) return false;

	// It does.  Queue the tokens for the main parse loop,
	// and tell it to retry...

	for (Iterator e = expn.iterator(); e.hasNext(); ) {
	    Token tok = (Token)e.next();
	    if (tok.type == Token.KEYWORD)
		deferredToks_.addFirst (new Token (cur_tok_, 
						   cur_tok_.indentLevel));
	    else
		deferredToks_.addFirst (new Token (tok, cur_tok_.indentLevel));
	}

	return true;
    }

    // Routine for debugging

    private void printDeferredToks() {
	for (java.util.Iterator it = deferredToks_.iterator(); it.hasNext();) 
	    System.err.println (it.next().toString());
    }

    // Dealing with other Xerces components.
    // Starting and ending elements.

    private void startElement (String element, XMLAttributes attrs)
	throws IOException, XNIException
    {
	contentHandler_.startElement (qnameize (element), attrs, augs);
    }

    private void endElement (String element) 
    {
	contentHandler_.endElement (qnameize (element), augs);
    }
	
    private QName qnameize (String str) {
	str = str.intern();
	return new QName (null, str, str, null);
    }

    // Document source methods.
    // Note that our DocumentHandler is initially the DTD validator,
    // but if it turns out that we aren't validating, it will call
    // this method to clip itself out of the processing pipeline.

    public void setDocumentHandler (XMLDocumentHandler handler) {
	contentHandler_ = handler;
    }

    public XMLDocumentHandler getDocumentHandler () {
	return contentHandler_;
    }

}
