//    XXML/Xerces Parser
//    Copyright (C) 2005 Robert S. Thau
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License verstion 2.1.
//
//    This library is distributed as a proof-of-concept only,
//    and WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package org.rsthau.xxml;

import java.io.IOException;

// Documented:

import org.apache.xerces.xni.XMLResourceIdentifier;
import org.apache.xerces.xni.XNIException;

import org.apache.xerces.xni.parser.XMLDTDScanner;
import org.apache.xerces.xni.parser.XMLInputSource;

// Probably documented:

import org.apache.xerces.util.XMLChar;
import org.apache.xerces.util.XMLResourceIdentifierImpl;

// Probably meant to be internal:

import org.apache.xerces.impl.XMLEntityManager;
import org.apache.xerces.impl.XMLEntityScanner;
import org.apache.xerces.impl.XMLEntityHandler;

class XxmlTokenizer 
    implements XMLEntityHandler
{
    private XMLEntityManager entityManager_;
    private XMLEntityScanner reader_;
    private XMLDTDScanner dtdScanner_;
    private char cur_char_;
    private boolean at_eof_ = false;

    public final static int EOF_INDENT = -1;

    public XxmlTokenizer (XMLEntityScanner r, 
			  XMLEntityManager em,
			  XMLDTDScanner dts) 
	throws IOException, SyntaxError
    {
	dtdScanner_ = dts;
	entityManager_ = em;
	reader_ = r;
	get_char();		// start the pipeline...
    }

    public boolean atEof() { 
	return at_eof_;
    }

    private Token nextRaw()
	throws IOException, SyntaxError
    {
	// Skip insignificant whitespace...

	while (!at_eof_ && XMLChar.isSpace (cur_char_))
	    get_char();

	if (at_eof_) { return new Token (Token.EOF, EOF_INDENT); }

	// Must now have a non-whitespace character.
	
	int pos = get_col_pos();

	switch (cur_char_) {
	case '=': get_char(); return new Token (Token.EQUALS, pos);
	case '{': get_char(); return new Token (Token.LBRACE, pos);
	case '}': get_char(); return new Token (Token.RBRACE, pos);
	case ';': get_char(); return new Token (Token.SEMICOLON, pos);
	case '"': get_char(); return scan_quoted_text(pos);
	case '#': get_char(); return scan_comment(pos);
	case '\\': get_char(); return new Token(Token.KEYWORD,scan_name(),
						pos);
	case ':': get_char(); 
	    if (XMLChar.isSpace (cur_char_))
		return new Token (Token.COLON,  pos);	    
	    else if (cur_char_ == ':') {
		get_char();
		return new Token (Token.NAME, ":" + scan_name(), pos);
	    }
	    else
		throw new SyntaxError ("':' may not be followed by '" +
				       cur_char_ + "'");

	default:
	    if (!XMLChar.isNameStart (cur_char_))
		throw new SyntaxError ("Unexpected character '" + 
				       cur_char_ + "'");
	    return new Token (Token.NAME, scan_name(), pos);
	}
    }

    public String scanBulkText (int endIndent)
	throws IOException, SyntaxError
    {
	// Accumulator for text.

	StringBuffer buf = new StringBuffer();

	// Read past the end of the line containing the \bulk_text token:

	while (!at_eof_ && cur_char_ != '\n') {
	    if (!XMLChar.isSpace (cur_char_))
		throw new SyntaxError ("Bulk text must begin on a new line");
	    get_char();
	}

	get_char();

	// Scan lines of bulk text.

	NEXT_LINE: while (!at_eof_) {

	    // Scan indent.

	    int cur_indent = get_col_pos();

	    while (!at_eof_ && cur_indent < endIndent) {
		get_char();
		if (cur_char_ == '\n') {
		    get_char();
		    buf.append ("\n");
		    continue NEXT_LINE;
		}
		else if (cur_char_ == ' ')
		    ++cur_indent;
		else
		    break NEXT_LINE; // we're done.
	    }

	    // Past indent.  Buffer up the rest of the line.

	    while (!at_eof_ && cur_char_ != '\n') {
		get_char();
		buf.append (cur_char_);
	    }
	}

	return buf.toString();
    }

    private Token scan_comment(int pos)
	throws IOException, SyntaxError
    {
	StringBuffer b = new StringBuffer();

	while (!at_eof_ && cur_char_ != '\n') {
	    b.append (cur_char_);
	    get_char();
	}

	return new Token (Token.COMMENT, b.toString().trim(), pos);
    }

    private Token scan_quoted_text (int pos)
	throws IOException, SyntaxError
    {
	StringBuffer b = new StringBuffer();

	while (!at_eof_) {
	    if (cur_char_ == '\n')
		throw new SyntaxError ("Newline in quoted string");
	    else if (cur_char_ == '"') {
		get_char();
		if (cur_char_ != '"') 
		    return new Token (Token.TEXT, b.toString(), pos);
	    }
	    b.append (cur_char_);
	    get_char();
	}

	throw new SyntaxError ("Unterminated quoted string");
    }

    private String scan_name ()
	throws IOException, SyntaxError
    {
	StringBuffer b = new StringBuffer();

	if (!XMLChar.isNameStart (cur_char_))
	    throw new SyntaxError ("No keyword following \\");

	while (XMLChar.isName (cur_char_) && cur_char_ != ':') {
	    b.append (cur_char_);
	    get_char();
	}
	return b.toString();
    }

    // Work around an apparent bug in the EntityScanner's
    // document location -- a bit awkward, as we don't want to
    // break if they fix the bug.

    private int charsOffset_ = 1;

    private int get_col_pos ()
    {
	return reader_.getColumnNumber() - charsOffset_;
    }

    private void get_char () 
	throws IOException, SyntaxError
    {
	assert !at_eof_;

	try {
	    cur_char_ = (char)reader_.scanChar();
	    if (cur_char_ == '\t')
		throw new SyntaxError ("Literal tab character not allowed");
	    if (cur_char_ == '\n')
		charsOffset_ = reader_.getColumnNumber();
	}
	catch (java.io.EOFException e) {
	    at_eof_ = true;
	    cur_char_ = ' ';	// terminates names;
	}
    }

    // Entity handler methods.  The only way these ever get called
    // is when an \include directive has set up a new entity.  So,
    // we generate ficticious tokens to mark the beginning and ending
    // of the #included file.
    //
    // We don't tell the contentHandler_ that we're starting or ending
    // a general entity -- if it cares at all, it will have expected
    // to see a declaration in the XML prolog, which we can't produce
    // for \include's of body content...

    private Token includes_deferred_tok_ = null;
    private Token includes_faked_tok_    = null;

    public Token next ()
	throws IOException, SyntaxError
    {
	if (includes_deferred_tok_ != null) {
	    Token it = includes_deferred_tok_;
	    includes_deferred_tok_ = null;
	    return it;
	}

	Token it = nextRaw();

	if (includes_faked_tok_ != null) {
	    includes_deferred_tok_ = it;
	    it = includes_faked_tok_;
	    includes_faked_tok_ = null;
	}

	return it;
    }

    public void startEntity (String name,
			     XMLResourceIdentifier identifier,
			     String encoding)
    {
	includes_faked_tok_ = new Token (Token.BEGIN_FILE,
					 identifier.getExpandedSystemId(),
					 XxmlTokenizer.EOF_INDENT);
    }

    public void endEntity (String name)
    {
	includes_faked_tok_ = new Token (Token.EOF,
					 XxmlTokenizer.EOF_INDENT);
    }

    // Reading in the DTD.  (The external subset, technically, but
    // since there's no syntax for an internal subset, it's the DTD).
    // Not an exception to the above rule:  the DTD scanner makes
    // *itself* the entity handler while the external entity containing
    // the DTD is being read. 
    // 
    // We aren't yet honoring the Xerces-specific property which tells
    // the parser to skip external DTDs.

    public void scanExternalDtd (String pubId, String sysId)
	throws IOException, XNIException
    {
	XMLResourceIdentifierImpl resId = 
	    new XMLResourceIdentifierImpl();

	resId.setValues (pubId, sysId, null, null);
	XMLInputSource inputSource = entityManager_.resolveEntity (resId);
	dtdScanner_.setInputSource (inputSource);

	while (dtdScanner_.scanDTDExternalSubset (true))
	    continue;

	entityManager_.setEntityHandler (this); 
	charsOffset_ = 1;	// Work around entity scanner positioning bugs
    }

}
